/*
 * Copyright (c) 1987, 1988, 1989, 1990, 1991 Stanford University
 * Copyright (c) 1991 Silicon Graphics, Inc.
 *
 * Permission to use, copy, modify, distribute, and sell this software and 
 * its documentation for any purpose is hereby granted without fee, provided
 * that (i) the above copyright notices and this permission notice appear in
 * all copies of the software and related documentation, and (ii) the names of
 * Stanford and Silicon Graphics may not be used in any advertising or
 * publicity relating to the software without the specific, prior written
 * permission of Stanford and Silicon Graphics.
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND, 
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  
 *
 * IN NO EVENT SHALL STANFORD OR SILICON GRAPHICS BE LIABLE FOR
 * ANY SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND,
 * OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS,
 * WHETHER OR NOT ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF 
 * LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE 
 * OF THIS SOFTWARE.
 */

#include "./memory.h"


/* kandrews 17 Feb 94 */
/* use memcpy, memset, etc.  instead of bcopy, bzero, etc. */

/* mpichler 18 Apr 94 */
/* use memmove instead of memcpy */



#include <memory.h>

#if defined(SUN4)
#include <stdlib.h>
#else if defined(SGI)
#include <string.h>
#endif


/*
 * Memory::copy should work also when memory regions overlap
 * memcpy(to, from, nbytes) does not guarantee this.
 * memmove guarantees this, memcpy does not.
 * On machines where memmove is not available, one should use
 * bcopy(from, to, nbytes);
 * which is also guaranteed to behave correctly.
 */

void Memory::copy(const void* from, void* to, unsigned int nbytes) {
#if defined(SUN4)
  bcopy(from, to, nbytes);
#else
  memmove(to, from, nbytes);
#endif
}


int Memory::compare(const void* b1, const void* b2, unsigned int nbytes) {
  return memcmp(b1, b2, nbytes) != 0;
}

void Memory::zero(void* b, unsigned int nbytes) {
  memset(b, 0, nbytes);
}
